import java.awt.*;
import ij.*;
import ij.gui.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.process.ShortProcessor;

/**
 *	This plugin simplifies the task of background subtraction and
 *	image normalization given a brightfield image and/or a background image.
 *	This plugin is currently limited to 16-bit images, but could easily be
 *	extended to 8-bit, 32-bit, or even color images. <p>
 *
 *	Possible normalization methods: <p><p>
 *	
 *	If both BRIGHT and DARK images are specified <p>
 *	  IMAGE = scale * (IMAGE - DARK) / (BRIGHT - DARK) <p>
 *  If only BRIGHT image is specified <p>
 *	  IMAGE = scale * IMAGE / BRIGHT <p>
 *	If only DARK image is specified <p>
 *	  IMAGE = IMAGE - DARK <p><p>
 *
 *	@author	Jeffrey Kuhn
 *	@author	The University of Texas at Austin
 *	@author	jkuhn@ccwf.cc.utexas.edu
*/
public class Normalize_ implements PlugInFilter {

	/** If "none" is selected, then that image is not used in the calculation */
	private static final String strNONE = "-- none --";

	/** Name of bright image */
	private static String strBrightName = null;		

	/** Name of dark image */
	private static String strDarkName = null;

	/** Number to scale resulting image by after division by BRIGHT */
	private static double dScale = 4095.0;			

	/** ImageJ ID of bright image */
	private int iBrightID;							

	/** ImageJ ID of dark image */
	private int iDarkID;		
	
	/** The image window to work on */					
	private ImagePlus impImage;

	/**
	 * Report what kind of images this plugin can handle to ImageJ
	 */
	public int setup(String arg, ImagePlus imp) {
		if (IJ.versionLessThan("1.18p"))
			return DONE;
		impImage = imp;
		return DOES_16;
		/* TODO: Support other image types */
	}

	/**
	 * Request the user to specify the bright and dark images
	 */
	public boolean getArguments() {
		int[] aiWinList = WindowManager.getIDList();
		if (aiWinList==null) {
			IJ.error("No windows are open.");
			return false;
		}
		String[] astrWinTitles = new String[aiWinList.length + 1];
		astrWinTitles[0] = strNONE;
		for (int i=0; i<aiWinList.length; i++) {
			ImagePlus imp = WindowManager.getImage(aiWinList[i]);
			if (imp != null) {
				astrWinTitles[i+1] = imp.getTitle();
			} else {
				astrWinTitles[i+1] = "";
			}
		}
		if (strBrightName == null)
			strBrightName = strNONE;
		if (strDarkName == null)
			strDarkName = strNONE;

		GenericDialog gd = new GenericDialog("Normalize Image", IJ.getInstance());

		gd.addMessage("IMAGE to normalize: " + impImage.getTitle());
		gd.addChoice("BRIGHT Image", astrWinTitles, strBrightName);
		gd.addChoice("DARK Image", astrWinTitles, strDarkName);
		gd.addNumericField("scale", dScale, 2);

		gd.addMessage("Possible normalization methods");
		gd.addMessage("= scale*(IMAGE-DARK)/(BRIGHT-DARK)");
		gd.addMessage("= scale*IMAGE/BRIGHT");
		gd.addMessage("= IMAGE-DARK");


		gd.showDialog();
		if (gd.wasCanceled()) {
			return false;
		}

		int iIndex;

		dScale = gd.getNextNumber();

		iIndex = gd.getNextChoiceIndex();
		if (iIndex == 0) {
			iBrightID = 0;
			strBrightName = null;
		} else {
			iBrightID = aiWinList[iIndex-1];
			strBrightName = astrWinTitles[iIndex];
		}

		iIndex = gd.getNextChoiceIndex();
		if (iIndex == 0) {
			iDarkID = 0;
			strDarkName = null;
		} else {
			iDarkID = aiWinList[iIndex-1];
			strDarkName = astrWinTitles[iIndex];
		}

		return true;
	}

	/**
	 * Normalize the current image.
	 */
	public void run(ImageProcessor ip) {
		// ask the user for the bright and dark images
		if (!getArguments()) {
			return;
		}

		int iW = ip.getWidth();
		int iH = ip.getHeight();
		int iLen = iW * iH;

		ImagePlus impBright = null;
		ImagePlus impDark = null;

		if (strBrightName != null) {
			impBright = WindowManager.getImage(iBrightID);
			if (impBright.getType() != impImage.getType()) {
				IJ.error("The BRIGHT image type does not match this image type.");
				return;
			}
			if ((impBright.getProcessor().getWidth() != iW) || (impBright.getProcessor().getHeight() != iH)) {
				IJ.error("The BRIGHT image size does not match this image size.");
				return;
			}
		}
		if (strDarkName != null) {
			impDark = WindowManager.getImage(iDarkID);
			if (impDark.getType() != impImage.getType()) {
				IJ.error("The DARK image type does not match this image type.");
				return;
			}
			if ((impDark.getProcessor().getWidth() != iW) || (impDark.getProcessor().getHeight() != iH)) {
				IJ.error("The DARK image size does not match this image size.");
				return;
			}
		}

		// At least one of bright or dark must have been specified
		if ((impBright == null) && (impDark == null)) {
			IJ.error("Neither a BRIGHT image nor a DARK image were specified. Normalization cannot proceed.");
			return;
		}

		switch (impImage.getType()) {
		case ImagePlus.GRAY16:
			ShortProcessor sp = (ShortProcessor)impImage.getProcessor();
				processGray16Unsigned(dScale, impImage, impBright, impDark, iLen);
			break;

		/* TODO: Support other image types */

		default:
			IJ.error("Normalize does not support this image type.");
			return;
		}

		// Reset the min and max displayed values
		double min = ip.getMin();
		double max = ip.getMax();
		ip.setMinAndMax(min, max);
	}

	/**
	 * Normalize a 16-bit unsigned image
	 */
	void processGray16Unsigned(double dScale, ImagePlus impImage, ImagePlus impBright, 
							 ImagePlus impDark, int iLen) {
		if ((impBright != null) && (impDark != null)) {
			short[] asImage = (short[])impImage.getProcessor().getPixels();
			short[] asBright = (short[])impBright.getProcessor().getPixels();
			short[] asDark = (short[])impDark.getProcessor().getPixels();
			for (int i=0; i<iLen; i++) {
				asImage[i] = (short)(dScale * ((asImage[i]&0xffff) - (asDark[i]&0xffff)) / ((asBright[i]&0xffff) - (asDark[i]&0xffff)));
			}
		} else if (impBright != null) {
			short[] asImage = (short[])impImage.getProcessor().getPixels();
			short[] asBright = (short[])impBright.getProcessor().getPixels();
			for (int i=0; i<iLen; i++) {
				asImage[i] = (short)(dScale * (asImage[i]&0xffff) / (asBright[i]&0xffff));
			}
		} else if (impDark != null) {
			short[] asImage = (short[])impImage.getProcessor().getPixels();
			short[] asDark = (short[])impDark.getProcessor().getPixels();
			for (int i=0; i<iLen; i++) {
				asImage[i] = (short)((asImage[i]&0xffff) - (asDark[i]&0xffff));
			}
		}
	}

	/* TODO: Support other image types */
}
