import ij.*;
import ij.plugin.filter.*;
import ij.process.*;
import ij.gui.*;
import ij.measure.*;
import ij.text.*;
import java.util.*;

/** 	Besides area, perimeter and angle, this plugin calculates shape descriptors. The mean and standard deviation of shape values
		are given for every image in a stack. The definitions are given according to Russ, 1999 (The image processing Handbook).
		The following descriptors are given:
			Form factor: 4pi*area/sqr(perimeter)
			Roundness: 4*area/pi*sqr(major axis)
			Compactness: sqrt((4/pi)*area)/major axis
			Aspect Ratio: major axis/minor axis

	The Analyze/Set Measurements... "Limit to Threshold" option must be checked (only thresholded pixels are included
	in measurement calculations). Use Image/Adjust/Threshold to set the threshold limits.
	Edge particles must also be excluded during measurement.

	Gary Chinga 020709

	Modification
	020829	Added option to display single particle shape descriptors

 */

public class Shape_Descriptors implements PlugInFilter, Measurements {
	ImagePlus imp;
	ResultsTable rt;
	int counter,slice,i,c;
 	TextWindow mtw;
	String title,headings,aLine ;
	float[] areas,per,mAxis,miAxis,angle,FF,Comp,Round,AR;
	float mArea, mPer,mMajor,mMinor,mAngle,mComp,mRound, mFF,mAR,ps;
	float sArea, sPer,sAngle,sComp,sRound, sFF,sAR;

	boolean canceled=false;
	private static String[] items = {"Single particle details","Mean particle details"};
	protected static final int SINGLE=0,MEAN=1;
	protected static int Choice;

	public int setup(String arg, ImagePlus imp) {
		if (IJ.versionLessThan("1.28"))
			return DONE;
		this.imp = imp;

		return DOES_8G;
	}

	public void run(ImageProcessor ip) {

        Calibration cal = imp.getCalibration();
        ps = (float) cal.pixelWidth;

		getDetails();
        if (canceled) return;

		int measurements = Analyzer.getMeasurements(); // defined in Set Measurements dialog
		Analyzer.setMeasurements(0);
		measurements |= AREA+PERIMETER+ELLIPSE; //make sure area and perimeter are measured
		Analyzer.setMeasurements(measurements);
		Analyzer a = new Analyzer();
		ParticleAnalyzer pa = new ParticleAnalyzer();
        pa.showDialog();

		int nSlices = imp.getStackSize();

		for (i=1; i<=nSlices; i++) {
			IJ.run("Clear Results");
			imp.setSlice(i);

			mArea = 0; mPer=0; mAngle=0; mComp=0; mRound=0; mFF=0;mAR=0;
			sArea = 0; sPer=0; sAngle=0; sComp=0; sRound=0; sFF=0;sAR=0;

			if (!pa.analyze(imp))
				return;

			rt =Analyzer.getResultsTable();  //get the system results table
			counter = rt.getCounter();
						//   	IJ.setColumnHeadings(rt.getColumnHeadings());
			if (counter>0) {
				areas = rt.getColumn(ResultsTable.AREA); // get area measurements
				per = rt.getColumn(ResultsTable.PERIMETER); // get perimeter measurements
				angle= rt.getColumn(ResultsTable.ANGLE); // get angle measurements
				mAxis= rt.getColumn(ResultsTable.MAJOR); // get area measurments
				miAxis= rt.getColumn(ResultsTable.MINOR); // get area measurments
				calculateShape();

				if (Choice==SINGLE) {
					for (int ii=0; ii<counter; ii++){
						c=ii+1;
						mArea=areas[ii];
						mPer=per[ii];
						mAngle=angle[ii];
					    mAR=AR[ii];
					    mComp=Comp[ii];
					    mRound=Round[ii];
        				mFF=FF[ii];
						writeResults();
					}

				}

				if (Choice==MEAN) {
					c=counter;
					mArea=calculateMean(areas);
					mPer=calculateMean(per);
					mAngle=calculateMean(angle);
					mFF=calculateMean(FF);
					mComp = calculateMean(Comp);
					mRound=calculateMean(Round);
					mAR=calculateMean(AR);

					sArea=calculateStd(mArea,areas);
					sPer=calculateStd(mPer,per);
					sAngle=calculateStd(mAngle,angle);
					sFF=calculateStd(mFF,FF);
					sComp = calculateStd(mComp,Comp);
					sRound=calculateStd(mRound,Round);
					sAR=calculateStd(mAR,AR);

					writeResults();
				}


			}
			else {c=0; writeResults();}
		}
	}

	void calculateShape(){
				FF = new float[areas.length];
				Round = new float[areas.length];
				Comp = new float[areas.length];
				AR = new float[areas.length];

				for (int ii=0; ii<counter; ii++){
					FF[ii] =(float) ((4*Math.PI*areas[ii])/(sqr(per[ii])));
					Round[ii] = (float) ((4*areas[ii])/(Math.PI*sqr(mAxis[ii])));
					Comp[ii] = (float) (Math.sqrt((4/Math.PI)*areas[ii])/mAxis[ii]);
					AR[ii] = (float) (mAxis[ii]/miAxis[ii]);
				}
	}

	float calculateMean(float[] dataset){
		double mValue=0;
		for (int j=0; j<counter; j++){mValue += dataset[j];}
		return (float) (mValue/counter);
	}

	float calculateStd(float mValue,float[] dataset){
		float sValue=0;
		if (counter==1) {return (float) (sValue);}
		else{
			for (int j=0; j<counter; j++){sValue += sqr(mValue-dataset[j]);}
			return (float) (Math.sqrt(sValue/(counter-1)));
		}
	}


	void writeResults(){
			if (Choice==MEAN)
				aLine = i+"\t"+c+"\t"+IJ.d2s(mArea,2)+"\t"+IJ.d2s(sArea,2)+"\t"+IJ.d2s(mPer,2)+"\t"+IJ.d2s(sPer,2)+"\t"+IJ.d2s(mAngle,2)+"\t"+IJ.d2s(sAngle,2)+"\t"+IJ.d2s(mAR,2)+"\t"+IJ.d2s(sAR,2)+"\t"+IJ.d2s(mComp,2)+"\t"+IJ.d2s(sComp,2)+"\t"+IJ.d2s(mRound,2)+"\t"+IJ.d2s(sRound,2)+"\t"+IJ.d2s(mFF,2)+"\t"+IJ.d2s(sFF,2);

            else aLine = i+"\t"+c+"\t"+IJ.d2s(mArea,2)+"\t"+IJ.d2s(mPer,2)+"\t"+IJ.d2s(mAngle,2)+"\t"+IJ.d2s(mAR,2)+"\t"+IJ.d2s(mComp,2)+"\t"+IJ.d2s(mRound,2)+"\t"+IJ.d2s(mFF,2);

            if (mtw==null) {
				if (Choice==MEAN){
          			title = "Mean values of "+imp.getShortTitle();
					headings = "Slice\tCount\tArea\tstd\tPerimeter\tstd\tAngle\tstd\tAR\tstd\tComp\tstd\tRound\tstd\tForm Factor\tstd";
					mtw = new TextWindow(title, headings, aLine, 850, 180);
				}
            	else {
					title = "Single particle values of "+imp.getShortTitle();
					headings = "Slice\tCount\tArea\tPerimeter\tAngle\tAR\tComp\tRound\tForm Factor";
            		mtw = new TextWindow(title, headings, aLine, 550, 180);
				}
            } else
           		mtw.append(aLine);
	}


	double sqr(double x) {return x*x;}

 	void getDetails() {
	       	GenericDialog gd = new GenericDialog("Particle details...");
	       	String units = imp.getCalibration().getUnits();
	    	gd.addChoice("Display", items, items[Choice]);
	    	if (!units.startsWith("pixel"))
				gd.addMessage("        (Pixel size = "+IJ.d2s(ps,2)+" "+units+")");
	       	gd.showDialog();
       		if (gd.wasCanceled()){
                        		canceled = true;
                        		return;
			}
			Choice = gd.getNextChoiceIndex();
        	}

}
